@@ -1,4 +1,5 @@
 /* -*- Mode: C++; tab-width: 2; indent-tabs-mode: nil; c-basic-offset: 2 -*- */
+/* vim: set ts=2 sw=2 et tw=80: */
 /* ***** BEGIN LICENSE BLOCK *****
  * Version: MPL 1.1/GPL 2.0/LGPL 2.1
  *
@@ -44,6 +45,7 @@
 #include "plstr.h"
 #include "nsCRT.h"
 #include "nsIURI.h"
+#include "nsIJARURI.h"
 #include "nsNetUtil.h"
 #include "nsJSPrincipals.h"
 #include "nsVoidArray.h"
@@ -113,6 +115,9 @@ nsPrincipal::Init(const nsACString& aCertFingerprint,
 
   mCodebase = aCodebase;
 
+  // Invalidate our cached origin
+  mOrigin = nsnull;
+
   nsresult rv;
   if (!aCertFingerprint.IsEmpty()) {
     rv = SetCertificate(aCertFingerprint, aSubjectName, aPrettyName, aCert);
@@ -163,10 +168,21 @@ nsPrincipal::GetOrigin(char **aOrigin)
 {
   *aOrigin = nsnull;
 
-  nsIURI* uri = mDomain ? mDomain : mCodebase;
-  if (!uri) {
-    NS_ASSERTION(mCert, "No Domain or Codebase for a non-cert principal");
-    return NS_ERROR_FAILURE;
+  if (!mOrigin) {
+    mOrigin = mDomain ? mDomain : mCodebase;
+    if (mOrigin) {
+      // If uri is a jar URI, get the base URI
+      nsCOMPtr<nsIJARURI> jarURI;
+      while((jarURI = do_QueryInterface(mOrigin)))
+      {
+        jarURI->GetJARFile(getter_AddRefs(mOrigin));
+      }
+    }
+  }
+      
+  if (!mOrigin) {
+      NS_ASSERTION(mCert, "No Domain or Codebase for a non-cert principal");
+      return NS_ERROR_FAILURE;
   }
 
   nsCAutoString hostPort;
@@ -176,14 +192,14 @@ nsPrincipal::GetOrigin(char **aOrigin)
   // XXX this should be removed in favor of the solution in
   // bug 160042.
   PRBool isChrome;
-  nsresult rv = uri->SchemeIs("chrome", &isChrome);
+  nsresult rv = mOrigin->SchemeIs("chrome", &isChrome);
   if (NS_SUCCEEDED(rv) && !isChrome) {
-    rv = uri->GetHostPort(hostPort);
+    rv = mOrigin->GetHostPort(hostPort);
   }
 
   if (NS_SUCCEEDED(rv) && !isChrome) {
     nsCAutoString scheme;
-    rv = uri->GetScheme(scheme);
+    rv = mOrigin->GetScheme(scheme);
     NS_ENSURE_SUCCESS(rv, rv);
     *aOrigin = ToNewCString(scheme + NS_LITERAL_CSTRING("://") + hostPort);
   }
@@ -191,7 +207,7 @@ nsPrincipal::GetOrigin(char **aOrigin)
     // Some URIs (e.g., nsSimpleURI) don't support host. Just
     // get the full spec.
     nsCAutoString spec;
-    rv = uri->GetSpec(spec);
+    rv = mOrigin->GetSpec(spec);
     NS_ENSURE_SUCCESS(rv, rv);
     *aOrigin = ToNewCString(spec);
   }
@@ -259,15 +275,10 @@ nsPrincipal::Equals(nsIPrincipal *aOther, PRBool *aResult)
     }
 
     // Codebases are equal if they have the same origin.
-    nsIURI *origin = mDomain ? mDomain : mCodebase;
-    nsCOMPtr<nsIURI> otherOrigin;
-    aOther->GetDomain(getter_AddRefs(otherOrigin));
-    if (!otherOrigin) {
-      aOther->GetURI(getter_AddRefs(otherOrigin));
-    }
-
-    return nsScriptSecurityManager::GetScriptSecurityManager()
-           ->SecurityCompareURIs(origin, otherOrigin, aResult);
+    *aResult =
+      NS_SUCCEEDED(nsScriptSecurityManager::GetScriptSecurityManager()
+                   ->CheckSameOriginPrincipal(this, aOther));
+    return NS_OK;
   }
 
   *aResult = PR_TRUE;
@@ -499,6 +510,9 @@ void
 nsPrincipal::SetURI(nsIURI* aURI)
 {
   mCodebase = aURI;
+
+  // Invalidate our cached origin
+  mOrigin = nsnull;
 }
 
 
@@ -597,6 +611,9 @@ nsPrincipal::SetDomain(nsIURI* aDomain)
   // Domain has changed, forget cached security policy
   SetSecurityPolicy(nsnull);
 
+  // Invalidate our cached origin
+  mOrigin = nsnull;
+
   return NS_OK;
 }
 
@@ -635,6 +652,9 @@ nsPrincipal::InitFromPersistent(const char* aPrefName,
     }
 
     mTrusted = aTrusted;
+
+    // Invalidate our cached origin
+    mOrigin = nsnull;
   }
 
   rv = mJSPrincipals.Init(this, aToken.get());
