@@ -103,10 +103,19 @@
 #include "nsIEventStateManager.h"
 #include "nsXFormsModelElement.h"
 
+#include "prtime.h"
+#include "nsIScriptGlobalObject.h"
+#include "nsIInterfaceRequestor.h"
+#include "nsIInterfaceRequestorUtils.h"
+#include "nsIPrompt.h"
+
+// For locale aware string methods
+#include "plstr.h"
+
 #define CANCELABLE 0x01
 #define BUBBLES    0x02
 
-const EventData sXFormsEventsEntries[42] = {
+const EventData sXFormsEventsEntries[43] = {
   { "xforms-model-construct",      PR_FALSE, PR_TRUE  },
   { "xforms-model-construct-done", PR_FALSE, PR_TRUE  },
   { "xforms-ready",                PR_FALSE, PR_TRUE  },
@@ -148,7 +157,8 @@ const EventData sXFormsEventsEntries[42] = {
   { "xforms-link-exception",       PR_FALSE, PR_TRUE  },
   { "xforms-link-error",           PR_FALSE, PR_TRUE  },
   { "xforms-compute-exception",    PR_FALSE, PR_TRUE  },
-  { "xforms-moz-hint-off",         PR_FALSE, PR_TRUE  }
+  { "xforms-moz-hint-off",         PR_FALSE, PR_TRUE  },
+  { "xforms-submit-serialize",     PR_FALSE, PR_TRUE  }
 };
 
 static const EventData sEventDefaultsEntries[] = {
@@ -204,8 +214,29 @@ struct EventItem
   nsXFormsEvent           event;
   nsCOMPtr<nsIDOMNode>    eventTarget;
   nsCOMPtr<nsIDOMElement> srcElement;
+  nsCOMArray<nsIXFormsContextInfo> *contextInfo;
 };
 
+PRInt32 nsXFormsUtils::waitLimit = 10;
+
+PR_STATIC_CALLBACK(int) PrefChangedCallback(const char* aPref, void* aData)
+{
+  nsresult rv;
+  nsCOMPtr<nsIPrefBranch> pref = do_GetService(NS_PREFSERVICE_CONTRACTID, &rv);
+  if (NS_SUCCEEDED(rv) && pref) {
+
+    // if our prefs changed, make sure to update our static variables.
+    if (strcmp(aPref, PREF_WAIT_LIMIT) == 0) {
+      PRInt32 val;
+      if (NS_SUCCEEDED(pref->GetIntPref(PREF_WAIT_LIMIT, &val))) {
+        nsXFormsUtils::waitLimit = val;
+      }
+    }
+  }
+
+  return 0; // PREF_OK
+}
+
 /* static */ nsresult
 nsXFormsUtils::Init()
 {
@@ -235,6 +266,39 @@ nsXFormsUtils::Init()
     sEventDefaults.Put(NS_ConvertUTF8toUTF16(sEventDefaultsEntries[i].name),
                                              flag);
   }
+
+  nsresult rv;
+  nsCOMPtr<nsIPrefBranch> prefBranch =
+    do_GetService(NS_PREFSERVICE_CONTRACTID, &rv);
+  if (NS_SUCCEEDED(rv) && prefBranch) {
+    PRInt32 intval;
+    rv = prefBranch->GetIntPref(PREF_WAIT_LIMIT, &intval);
+    if (NS_SUCCEEDED(rv)) {
+      nsXFormsUtils::waitLimit = intval;
+    }
+  }
+
+  nsCOMPtr<nsIPref> pref = do_GetService(NS_PREF_CONTRACTID, &rv);
+  NS_ENSURE_STATE(pref);
+  rv = pref->RegisterCallback(PREF_WAIT_LIMIT,
+                              PrefChangedCallback, nsnull);
+  NS_ENSURE_SUCCESS(rv, rv);
+
+  return NS_OK;
+}
+
+/* static */ nsresult
+nsXFormsUtils::Shutdown()
+{
+  // unregister our pref listeners
+
+  nsresult rv;
+  nsCOMPtr<nsIPref> pref = do_GetService(NS_PREF_CONTRACTID, &rv);
+  NS_ENSURE_STATE(pref);
+  rv = pref->UnregisterCallback(PREF_WAIT_LIMIT,
+                                PrefChangedCallback, nsnull);
+  NS_ENSURE_SUCCESS(rv, rv);
+
   return NS_OK;
 }
 
@@ -290,7 +354,8 @@ nsXFormsUtils::GetNodeContext(nsIDOMElement           *aElement,
                               nsIXFormsControl       **aParentControl,
                               nsIDOMNode             **aContextNode,
                               PRInt32                 *aContextPosition,
-                              PRInt32                 *aContextSize)
+                              PRInt32                 *aContextSize,
+                              PRBool                   aUseBindAttr)
 {
   NS_ENSURE_ARG(aElement);
   NS_ENSURE_ARG(aOuterBind);
@@ -310,7 +375,7 @@ nsXFormsUtils::GetNodeContext(nsIDOMElement           *aElement,
   nsAutoString bindId;
   NS_NAMED_LITERAL_STRING(bindStr, "bind");
   aElement->GetAttribute(bindStr, bindId);
-  if (!bindId.IsEmpty()) {
+  if (!bindId.IsEmpty() && aUseBindAttr) {
     // CASE 1: Use @bind
     GetElementByContextId(aElement, bindId, aBindElement);
 
@@ -438,7 +503,7 @@ nsXFormsUtils::EvaluateXPath(const nsAString        &aExpression,
   NS_ENSURE_STATE(eval);
 
   nsCOMPtr<nsIDOMNSXPathExpression> expression;
-  nsresult rv = eval->CreateExpression(aExpression, aResolverNode,
+  nsresult rv = eval->CreateExpression(aExpression, aResolverNode, aContextNode,
                          getter_AddRefs(expression));
   PRBool throwException = PR_FALSE;
   if (!expression) {
@@ -462,7 +527,7 @@ nsXFormsUtils::EvaluateXPath(const nsAString        &aExpression,
       /// @see http://bugzilla.mozilla.org/show_bug.cgi?id=265212
       if (aSet) {
         nsXFormsXPathParser parser;
-        nsXFormsXPathAnalyzer analyzer(eval, aResolverNode);
+        nsXFormsXPathAnalyzer analyzer(eval, aResolverNode, aContextNode);
         nsAutoPtr<nsXFormsXPathNode> xNode(parser.Parse(aExpression));
         rv = analyzer.Analyze(aContextNode,
                               xNode,
@@ -496,7 +561,21 @@ nsXFormsUtils::EvaluateXPath(const nsAString        &aExpression,
     nsCOMPtr<nsIDOMElement> resolverElement = do_QueryInterface(aResolverNode);
     nsCOMPtr<nsIModelElementPrivate> modelPriv = nsXFormsUtils::GetModel(resolverElement);
     nsCOMPtr<nsIDOMNode> model = do_QueryInterface(modelPriv);
-    DispatchEvent(model, eEvent_ComputeException, nsnull, resolverElement);
+
+    // Context Info: 'error-message'
+    // Error message containing the expression being processed.
+    nsAutoString errorMsg;
+    errorMsg.AssignLiteral("Error evaluating expression: ");
+    errorMsg.Append(aExpression);
+
+    nsCOMPtr<nsXFormsContextInfo> contextInfo =
+      new nsXFormsContextInfo(resolverElement);
+    NS_ENSURE_TRUE(contextInfo, NS_ERROR_OUT_OF_MEMORY);
+    contextInfo->SetStringValue("error-message", errorMsg);
+    nsCOMArray<nsIXFormsContextInfo> contextInfoArray;
+    contextInfoArray.AppendObject(contextInfo);
+    DispatchEvent(model, eEvent_ComputeException, nsnull, resolverElement,
+                  &contextInfoArray);
   }
 
   return rv;
@@ -807,7 +886,8 @@ nsXFormsUtils::GetSingleNodeBindingValue(nsIDOMElement* aElement,
 
 nsresult
 DispatchXFormsEvent(nsIDOMNode* aTarget, nsXFormsEvent aEvent,
-                    PRBool *aDefaultActionEnabled)
+                    PRBool *aDefaultActionEnabled,
+                    nsCOMArray<nsIXFormsContextInfo> *aContextInfo)
 {
   nsCOMPtr<nsIDOMDocument> domDoc;
   aTarget->GetOwnerDocument(getter_AddRefs(domDoc));
@@ -827,8 +907,12 @@ DispatchXFormsEvent(nsIDOMNode* aTarget, nsXFormsEvent aEvent,
 
   nsXFormsUtils::SetEventTrusted(event, aTarget);
 
+  // Create an nsXFormsDOMEvent with the context info for the event.
+  nsCOMPtr<nsIXFormsDOMEvent> xfDOMEvent = new nsXFormsDOMEvent(event, aContextInfo);
+  NS_ENSURE_TRUE(xfDOMEvent, NS_ERROR_OUT_OF_MEMORY);
+
   PRBool defaultActionEnabled = PR_TRUE;
-  nsresult rv = target->DispatchEvent(event, &defaultActionEnabled);
+  nsresult rv = target->DispatchEvent(xfDOMEvent, &defaultActionEnabled);
 
   if (NS_SUCCEEDED(rv) && aDefaultActionEnabled)
     *aDefaultActionEnabled = defaultActionEnabled;
@@ -860,6 +944,10 @@ DispatchXFormsEvent(nsIDOMNode* aTarget, nsXFormsEvent aEvent,
     break;
   }
 
+  // Clear the context info array before the next event.
+  if (aContextInfo)
+    aContextInfo->Clear();
+
   return rv;
 }
 
@@ -882,7 +970,8 @@ DeleteVoidArray(void    *aObject,
 
 nsresult
 DeferDispatchEvent(nsIDOMNode* aTarget, nsXFormsEvent aEvent,
-                   nsIDOMElement *aSrcElement)
+                   nsIDOMElement *aSrcElement,
+                   nsCOMArray<nsIXFormsContextInfo> *aContextInfo)
 {
   nsCOMPtr<nsIDOMDocument> domDoc;
   if (aTarget) {
@@ -912,6 +1001,7 @@ DeferDispatchEvent(nsIDOMNode* aTarget, nsXFormsEvent aEvent,
   deferredEvent->event = aEvent;
   deferredEvent->eventTarget = aTarget;
   deferredEvent->srcElement = aSrcElement;
+  deferredEvent->contextInfo = aContextInfo;
   eventList->AppendElement(deferredEvent);
 
   return NS_OK;
@@ -920,7 +1010,8 @@ DeferDispatchEvent(nsIDOMNode* aTarget, nsXFormsEvent aEvent,
 /* static */ nsresult
 nsXFormsUtils::DispatchEvent(nsIDOMNode* aTarget, nsXFormsEvent aEvent,
                              PRBool *aDefaultActionEnabled,
-                             nsIDOMElement *aSrcElement)
+                             nsIDOMElement *aSrcElement,
+                             nsCOMArray<nsIXFormsContextInfo> *aContextInfo)
 {
   // it is valid to have aTarget be null if this is an event that must be
   // targeted at a model per spec and aSrcElement is non-null.  Basically we
@@ -974,7 +1065,7 @@ nsXFormsUtils::DispatchEvent(nsIDOMNode* aTarget, nsXFormsEvent aEvent,
         // model later when the DOM is finished loading
         if (!aTarget) {
           if (aSrcElement) {
-            DeferDispatchEvent(aTarget, aEvent, aSrcElement);
+            DeferDispatchEvent(aTarget, aEvent, aSrcElement, aContextInfo);
             return NS_OK;
           }
           return NS_ERROR_FAILURE;
@@ -986,7 +1077,7 @@ nsXFormsUtils::DispatchEvent(nsIDOMNode* aTarget, nsXFormsEvent aEvent,
         PRBool safeToSendEvent = PR_FALSE;
         modelPriv->GetHasDOMContentFired(&safeToSendEvent);
         if (!safeToSendEvent) {
-          DeferDispatchEvent(aTarget, aEvent, nsnull);
+          DeferDispatchEvent(aTarget, aEvent, nsnull, aContextInfo);
           return NS_OK;
         }
       
@@ -1028,7 +1119,7 @@ nsXFormsUtils::DispatchEvent(nsIDOMNode* aTarget, nsXFormsEvent aEvent,
         PRBool safeToSendEvent = PR_FALSE;
         modelPriv->GetHasDOMContentFired(&safeToSendEvent);
         if (!safeToSendEvent) {
-          DeferDispatchEvent(aTarget, aEvent, nsnull);
+          DeferDispatchEvent(aTarget, aEvent, nsnull, aContextInfo);
           return NS_OK;
         }
       
@@ -1038,7 +1129,7 @@ nsXFormsUtils::DispatchEvent(nsIDOMNode* aTarget, nsXFormsEvent aEvent,
       break;
   }
 
-  return DispatchXFormsEvent(aTarget, aEvent, aDefaultActionEnabled);
+  return DispatchXFormsEvent(aTarget, aEvent, aDefaultActionEnabled, aContextInfo);
 
 }
 
@@ -1094,7 +1185,7 @@ nsXFormsUtils::DispatchDeferredEvents(nsIDOMDocument* aDocument)
       NS_ENSURE_STATE(item->eventTarget);
     }
 
-    DispatchXFormsEvent(item->eventTarget, item->event, nsnull);
+    DispatchXFormsEvent(item->eventTarget, item->event, nsnull, item->contextInfo);
   }
 
   doc->DeleteProperty(nsXFormsAtoms::deferredEventListProperty);
@@ -2388,3 +2479,62 @@ nsXFormsUtils::NodeHasItemset(nsIDOMNode *aNode)
   return hasItemset;
 }
 
+/* static */ PRBool
+nsXFormsUtils::AskStopWaiting(nsIDOMElement *aElement)
+{
+  nsCOMPtr<nsIDOMDocument> domDoc;
+  aElement->GetOwnerDocument(getter_AddRefs(domDoc));
+  nsCOMPtr<nsIDocument> doc(do_QueryInterface(domDoc));
+  NS_ENSURE_TRUE(doc, PR_TRUE);
+
+  nsCOMPtr<nsPIDOMWindow> win = doc->GetWindow();
+  NS_ENSURE_TRUE(win, PR_TRUE);
+
+  nsCOMPtr<nsIPrompt> prompt;
+  win->GetPrompter(getter_AddRefs(prompt));
+  NS_ENSURE_TRUE(prompt, PR_TRUE);
+
+  // Get localizable strings
+  nsCOMPtr<nsIStringBundleService>
+    stringService(do_GetService(NS_STRINGBUNDLE_CONTRACTID));
+  NS_ENSURE_TRUE(stringService, PR_TRUE);
+
+  nsCOMPtr<nsIStringBundle> bundle;
+  stringService->CreateBundle("chrome://global/locale/dom/dom.properties",
+                              getter_AddRefs(bundle));
+  NS_ENSURE_TRUE(bundle, PR_TRUE);
+  
+  nsXPIDLString title, msg, stopButton, waitButton;
+
+  nsresult rv;
+  rv = bundle->GetStringFromName(NS_LITERAL_STRING("KillScriptTitle").get(),
+                                 getter_Copies(title));
+  rv |= bundle->GetStringFromName(NS_LITERAL_STRING("StopScriptButton").get(),
+                                  getter_Copies(stopButton));
+  rv |= bundle->GetStringFromName(NS_LITERAL_STRING("WaitForScriptButton").get(),
+                                  getter_Copies(waitButton));
+  rv |= bundle->GetStringFromName(NS_LITERAL_STRING("KillScriptMessage").get(),
+                                  getter_Copies(msg));
+
+
+  // GetStringFromName can return NS_OK and still give NULL string
+  if (NS_FAILED(rv) || !title || !msg || !stopButton || !waitButton) {
+    NS_ERROR("Failed to get localized strings.");
+    return PR_TRUE;
+  }
+
+  PRInt32 buttonPressed = 1; // In case user exits dialog by clicking X
+  PRUint32 buttonFlags = (nsIPrompt::BUTTON_TITLE_IS_STRING *
+                          (nsIPrompt::BUTTON_POS_0 + nsIPrompt::BUTTON_POS_1));
+
+  // Open the dialog.
+  rv = prompt->ConfirmEx(title, msg, buttonFlags, stopButton, waitButton,
+                         nsnull, nsnull, nsnull, &buttonPressed);
+
+  if (NS_SUCCEEDED(rv) && (buttonPressed == 0)) {
+    // Request stopping the unresponsive code
+    return PR_TRUE;
+  } else {
+    return PR_FALSE;
+  }
+}
