@@ -53,7 +53,6 @@
 #include "nsISmtpUrl.h"
 #include "nsIURI.h"
 #include "nsMsgI18N.h"
-#include "nsIMsgDraft.h"
 #include "nsIMsgComposeParams.h"
 #include "nsXPCOM.h"
 #include "nsISupportsPrimitives.h"
@@ -76,6 +75,7 @@
 #include "nsIPrefBranch2.h"
 #include "nsMsgBaseCID.h"
 #include "nsIMsgAccountManager.h"
+#include "nsIMimeMiscStatus.h"
 
 #include "nsIInterfaceRequestorUtils.h"
 
@@ -92,6 +92,12 @@
 #include "nsIContentSink.h"
 #include "mozISanitizingSerializer.h"
 
+#include "nsMsgMimeCID.h"
+#include "nsNetUtil.h"
+#include "nsIMsgMailNewsUrl.h"
+#include "nsIMimeStreamConverter.h"
+#include "nsIStreamConverter.h"
+
 #ifdef MOZ_XUL_APP
 #include "nsICommandLine.h"
 #endif
@@ -175,6 +181,8 @@ nsMsgComposeService::~nsMsgComposeService()
     DeleteCachedWindows();
     delete [] mCachedWindows;
   }
+
+  mOpenComposeWindows.Clear();
 }
 
 nsresult nsMsgComposeService::Init()
@@ -195,10 +203,10 @@ nsresult nsMsgComposeService::Init()
   if (pbi)
     rv = pbi->AddObserver(PREF_MAIL_COMPOSE_MAXRECYCLEDWINDOWS, this, PR_TRUE);
 
+  mOpenComposeWindows.Init();
   Reset();
 
   AddGlobalHtmlDomains();
-	
   return rv;
 }
 
@@ -214,6 +222,8 @@ void nsMsgComposeService::Reset()
     mMaxRecycledWindows = 0;
   }
 
+  mOpenComposeWindows.Clear();
+
   nsCOMPtr<nsIPrefBranch> prefs(do_GetService(NS_PREFSERVICE_CONTRACTID));
   if(prefs)
     rv = prefs->GetIntPref(PREF_MAIL_COMPOSE_MAXRECYCLEDWINDOWS, &mMaxRecycledWindows);
@@ -414,26 +424,13 @@ nsMsgComposeService::OpenComposeWindow(const char *msgComposeWindowURL, const ch
   if (type == nsIMsgCompType::ForwardInline || type == nsIMsgCompType::Draft || type == nsIMsgCompType::Template
     || type == nsIMsgCompType::ReplyWithTemplate)
   {
-    nsCOMPtr<nsIMsgDraft> pMsgDraft (do_CreateInstance(NS_MSGDRAFT_CONTRACTID, &rv));
-    if (NS_SUCCEEDED(rv) && pMsgDraft)
-    {
       nsCAutoString uriToOpen(originalMsgURI);
-      uriToOpen.Append("?fetchCompleteMessage=true"); 
+      uriToOpen += (uriToOpen.FindChar('?') == kNotFound) ? "?" : "&";
+      uriToOpen.Append("fetchCompleteMessage=true"); 
 
-      switch(type)
-      {
-        case nsIMsgCompType::ForwardInline:
-            rv = pMsgDraft->OpenDraftMsg(uriToOpen.get(), originalMsgURI, identity, PR_TRUE, aMsgWindow);
-          break;
-        case nsIMsgCompType::Draft:
-            rv = pMsgDraft->OpenDraftMsg(uriToOpen.get(), nsnull, identity, PR_FALSE, aMsgWindow);
-          break;
-        case nsIMsgCompType::Template:
-            rv = pMsgDraft->OpenEditorTemplate(uriToOpen.get(), identity, aMsgWindow);
-          break;
-      }
-    }
-    return rv;
+      return LoadDraftOrTemplate(uriToOpen, type == nsIMsgCompType::ForwardInline || type == nsIMsgCompType::Draft ? 
+                                 nsMimeOutput::nsMimeMessageDraftOrTemplate : nsMimeOutput::nsMimeMessageEditorTemplate,  
+                                 identity, originalMsgURI, type == nsIMsgCompType::ForwardInline, aMsgWindow);
   }
 
   nsCOMPtr<nsIMsgComposeParams> pMsgComposeParams (do_CreateInstance(NS_MSGCOMPOSEPARAMS_CONTRACTID, &rv));
@@ -472,7 +469,23 @@ nsMsgComposeService::OpenComposeWindow(const char *msgComposeWindowURL, const ch
           pMsgCompFields->SetNewshost(host.get());
         }
         else
+        {
           pMsgComposeParams->SetOriginalMsgURI(originalMsgURI);
+          if (PL_strstr(originalMsgURI, "type=application/x-message-display"))
+          {
+            nsCOMPtr <nsIMsgDBHdr> msgHdr;
+            if (strncmp(originalMsgURI, "file:", 5))
+              rv = GetMsgDBHdrFromURI(originalMsgURI, getter_AddRefs(msgHdr));
+            if (aMsgWindow && !msgHdr)
+            {
+              nsCOMPtr <nsIMsgHeaderSink> headerSink;
+              rv = aMsgWindow->GetMsgHeaderSink(getter_AddRefs(headerSink));
+              if (headerSink)
+                rv = headerSink->GetDummyMsgHeader(getter_AddRefs(msgHdr));
+            }
+            pMsgComposeParams->SetOrigMsgHdr(msgHdr);
+          }
+        }
       }
 
       pMsgComposeParams->SetComposeFields(pMsgCompFields);
@@ -1339,6 +1352,129 @@ nsresult nsMsgComposeService::AddGlobalHtmlDomains()
   return NS_OK;
 }
 
+NS_IMETHODIMP
+nsMsgComposeService::RegisterComposeWindow(nsIDOMWindowInternal * aWindow, nsIMsgCompose *aComposeObject)
+{
+  NS_ENSURE_ARG_POINTER(aWindow);
+  NS_ENSURE_ARG_POINTER(aComposeObject);
+
+  nsresult rv;
+
+  // add the msg compose / dom window mapping to our hash table
+  nsCOMPtr<nsIWeakReference> weakDOMWindow = do_GetWeakReference(aWindow, &rv);
+  NS_ENSURE_SUCCESS(rv,rv);
+  nsCOMPtr<nsIWeakReference> weakMsgComposePtr = do_GetWeakReference(aComposeObject);
+  NS_ENSURE_SUCCESS(rv,rv);
+  mOpenComposeWindows.Put(weakDOMWindow, weakMsgComposePtr);
+  
+  return rv;
+}
+
+NS_IMETHODIMP
+nsMsgComposeService::UnregisterComposeWindow(nsIDOMWindowInternal * aWindow)
+{
+  NS_ENSURE_ARG_POINTER(aWindow);
+  
+  nsresult rv;
+  nsCOMPtr<nsIWeakReference> weakDOMWindow = do_GetWeakReference(aWindow, &rv);
+  NS_ENSURE_SUCCESS(rv,rv);
+
+  mOpenComposeWindows.Remove(weakDOMWindow);
+
+  return rv;
+}
+
+NS_IMETHODIMP
+nsMsgComposeService::GetMsgComposeForWindow(nsIDOMWindowInternal * aWindow, nsIMsgCompose ** aComposeObject)
+{
+  NS_ENSURE_ARG_POINTER(aWindow);
+  NS_ENSURE_ARG_POINTER(aComposeObject);
+
+  // get the weak reference for our dom window
+  nsresult rv;
+  nsCOMPtr<nsIWeakReference> weakDOMWindow = do_GetWeakReference(aWindow, &rv);
+  NS_ENSURE_SUCCESS(rv,rv);
+
+  nsCOMPtr<nsIWeakReference> weakMsgComposePtr;
+
+  NS_ENSURE_TRUE(mOpenComposeWindows.Get(weakDOMWindow, getter_AddRefs(weakMsgComposePtr)), NS_ERROR_FAILURE);
+
+  nsCOMPtr<nsIMsgCompose> msgCompose = do_QueryReferent(weakMsgComposePtr, &rv);
+  NS_ENSURE_SUCCESS(rv, rv);
+  
+  NS_IF_ADDREF(*aComposeObject = msgCompose);
+  return rv;
+}
+
+/**
+ * LoadDraftOrTemplate
+ *   Helper routine used to run msgURI through libmime in order to fetch the contents for a
+ *   draft or template.
+ */
+nsresult    
+nsMsgComposeService::LoadDraftOrTemplate(const nsACString& aMsgURI, nsMimeOutputType aOutType, 
+                                         nsIMsgIdentity * aIdentity, const char * aOriginalMsgURI, 
+                                         PRBool aAddInlineHeaders, nsIMsgWindow *aMsgWindow)
+{
+  nsresult rv;
+  nsCOMPtr <nsIMsgMessageService> messageService;
+  rv = GetMessageServiceFromURI(nsPromiseFlatCString(aMsgURI).get(), getter_AddRefs(messageService));
+  NS_ENSURE_SUCCESS(rv, rv);
+
+  // Now, we can create a mime parser (nsIStreamConverter)!
+  nsCOMPtr<nsIMimeStreamConverter> mimeConverter = 
+    do_CreateInstance(NS_MAILNEWS_MIME_STREAM_CONVERTER_CONTRACTID, &rv);
+  NS_ENSURE_SUCCESS(rv, rv);
+  
+  mimeConverter->SetMimeOutputType(aOutType);  // Set the type of output for libmime
+  mimeConverter->SetForwardInline(aAddInlineHeaders);
+  mimeConverter->SetIdentity(aIdentity);
+  mimeConverter->SetOriginalMsgURI(aOriginalMsgURI);
+
+  nsCOMPtr<nsIURI> url;
+  PRBool fileUrl = StringBeginsWith(aMsgURI, NS_LITERAL_CSTRING("file:"));
+  nsACString::const_iterator start, end;
+  aMsgURI.BeginReading(start);
+  aMsgURI.EndReading(end);
+  if (fileUrl || FindInReadable(NS_LITERAL_CSTRING("&type=application/x-message-display"), start, end))
+    rv = NS_NewURI(getter_AddRefs(url), aMsgURI);
+  else
+    rv = messageService->GetUrlForUri(nsPromiseFlatCString(aMsgURI).get(), getter_AddRefs(url), aMsgWindow);
+  NS_ENSURE_SUCCESS(rv, rv);
+
+  rv = url->SetSpec(aMsgURI);
+  NS_ENSURE_SUCCESS(rv, rv);
+
+  // if we are forwarding a message and that message used a charset over ride
+  // then use that over ride charset instead of the charset specified in the message
+  nsXPIDLCString mailCharset;
+  if (aMsgWindow)
+  {
+    PRBool charsetOverride;
+    if (NS_SUCCEEDED(aMsgWindow->GetCharsetOverride(&charsetOverride)) && charsetOverride)
+    {
+      if (NS_SUCCEEDED(aMsgWindow->GetMailCharacterSet(getter_Copies(mailCharset))))
+      {
+        nsCOMPtr<nsIMsgI18NUrl> i18nUrl(do_QueryInterface(url));
+        if (i18nUrl)
+          (void) i18nUrl->SetCharsetOverRide(mailCharset.get());
+      }
+    }
+  }
+
+  nsCOMPtr<nsIChannel> channel;
+  rv = NS_NewInputStreamChannel(getter_AddRefs(channel), url, nsnull);
+  NS_ENSURE_SUCCESS(rv, rv);
+  
+  nsCOMPtr<nsIStreamConverter> converter = do_QueryInterface(mimeConverter);
+  rv = converter->AsyncConvertData(nsnull, nsnull, nsnull, channel);
+  NS_ENSURE_SUCCESS(rv, rv);
+
+  // Now, just plug the two together and get the hell out of the way!
+  nsCOMPtr<nsIStreamListener> streamListener = do_QueryInterface(mimeConverter); 
+  return messageService->DisplayMessage(nsPromiseFlatCString(aMsgURI).get(), streamListener, aMsgWindow, nsnull, mailCharset, nsnull);;
+}
+
 #ifdef MOZ_XUL_APP
 NS_IMETHODIMP
 nsMsgComposeService::Handle(nsICommandLine* aCmdLine)
@@ -1346,6 +1482,7 @@ nsMsgComposeService::Handle(nsICommandLine* aCmdLine)
   nsresult rv;
   PRInt32 found, end, count;
   nsAutoString uristr;
+  PRBool composeShouldHandle = PR_TRUE;
 
   rv = aCmdLine->FindFlag(NS_LITERAL_STRING("compose"), PR_FALSE, &found);
   NS_ENSURE_SUCCESS(rv, rv);
@@ -1353,7 +1490,12 @@ nsMsgComposeService::Handle(nsICommandLine* aCmdLine)
   // MAC OS X passes in -url mailto:mscott@mozilla.org into the command line
   // instead of -compose.
   if (found == -1)
+  {
     rv = aCmdLine->FindFlag(NS_LITERAL_STRING("url"), PR_FALSE, &found);
+    // we don't want to consume the argument for -url unless we're sure it is a mailto url and we'll
+    // figure that out shortly.
+    composeShouldHandle = PR_FALSE;
+  }
 
   if (found == -1)
     return NS_OK;
@@ -1363,9 +1505,15 @@ nsMsgComposeService::Handle(nsICommandLine* aCmdLine)
   rv = aCmdLine->GetLength(&count);
   NS_ENSURE_SUCCESS(rv, rv);
 
-  if (count >= found) {
+  if (count > found + 1) {
     aCmdLine->GetArgument(found + 1, uristr);
-    if (StringBeginsWith(uristr, NS_LITERAL_STRING("mailto:"))) {
+    if (StringBeginsWith(uristr, NS_LITERAL_STRING("mailto:"))  ||
+        StringBeginsWith(uristr, NS_LITERAL_STRING("to="))  ||
+        StringBeginsWith(uristr, NS_LITERAL_STRING("cc="))  ||
+        StringBeginsWith(uristr, NS_LITERAL_STRING("subject=")) ||
+        StringBeginsWith(uristr, NS_LITERAL_STRING("body="))  ||
+        StringBeginsWith(uristr, NS_LITERAL_STRING("attachment="))) {
+      composeShouldHandle = PR_TRUE; // the -url argument looks like mailto
       end++;
       // mailto: URIs are frequently passed with spaces in them. They should be
       // escaped with %20, but we hack around broken clients. See bug 231032.
@@ -1384,24 +1532,26 @@ nsMsgComposeService::Handle(nsICommandLine* aCmdLine)
       uristr.Truncate();
     }
   }
+  if (composeShouldHandle)
+  {
+    aCmdLine->RemoveArguments(found, end);
 
-  aCmdLine->RemoveArguments(found, end);
+    nsCOMPtr<nsIWindowWatcher> wwatch (do_GetService(NS_WINDOWWATCHER_CONTRACTID));
+    NS_ENSURE_TRUE(wwatch, NS_ERROR_FAILURE);
 
-  nsCOMPtr<nsIWindowWatcher> wwatch (do_GetService(NS_WINDOWWATCHER_CONTRACTID));
-  NS_ENSURE_TRUE(wwatch, NS_ERROR_FAILURE);
+    nsCOMPtr<nsISupportsString> arg;
+    if (!uristr.IsEmpty()) {
+      arg = do_CreateInstance(NS_SUPPORTS_STRING_CONTRACTID);
+      if (arg)
+        arg->SetData(uristr);
+    }    
 
-  nsCOMPtr<nsISupportsString> arg;
-  if (!uristr.IsEmpty()) {
-    arg = do_CreateInstance(NS_SUPPORTS_STRING_CONTRACTID);
-    if (arg)
-      arg->SetData(uristr);
-  }    
+    nsCOMPtr<nsIDOMWindow> opened;
+    wwatch->OpenWindow(nsnull, DEFAULT_CHROME, "_blank",
+                       "chrome,dialog=no,all", arg, getter_AddRefs(opened));
 
-  nsCOMPtr<nsIDOMWindow> opened;
-  wwatch->OpenWindow(nsnull, DEFAULT_CHROME, "_blank",
-                     "chrome,dialog=no,all", arg, getter_AddRefs(opened));
-
-  aCmdLine->SetPreventDefault(PR_TRUE);
+    aCmdLine->SetPreventDefault(PR_TRUE);
+  }
   return NS_OK;
 }
 
